package PScm::SpecialForm;

use strict;
use warnings;
use base qw(PScm::Expr);

###############################
package PScm::SpecialForm::Let;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    my ($symbols, $values, $body) = $self->UnPack($form);

    $env->Extend(
        $symbols, $values,
        cont {
            my ($newenv, $fail) = @_;
            $body->Eval($newenv, $cont, $fail);
        },
        $fail
    );
}

sub UnPack {
    my ($self, $form) = @_;

    my ($bindings, $body) = $form->value;
    my ($symbols, $values) = $self->unpack_bindings($bindings);
    return ($symbols, $values, $body);
}

sub unpack_bindings {
    my ($self, $bindings) = @_;
    if ($bindings->is_null) {
        my $null = new PScm::Expr::List::Null();
        return ($null, $null);
    } else {
        my ($symbols, $values) =
            $self->unpack_bindings($bindings->rest);
        return (
            PScm::Expr::List->Cons($bindings->first->first,
                                   $symbols),
            PScm::Expr::List->Cons($bindings->first->rest->first,
                                   $values)
        );
    }
}

##################################
package PScm::SpecialForm::LetRec;

use base qw(PScm::SpecialForm::Let);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    my ($symbols, $values, $body) = $self->UnPack($form);

    $env->ExtendRecursively(
        $symbols, $values,
        cont {
            my ($extended_env, $fail) = @_;
            $body->Eval($extended_env, $cont, $fail);
        },
        $fail
    );
}

###################################
package PScm::SpecialForm::LetStar;

use base qw(PScm::SpecialForm::Let);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    my ($symbols, $values, $body) = $self->UnPack($form);

    $env->ExtendIteratively(
        $symbols, $values,
        cont {
            my ($extended_env, $fail) = @_;
            $body->Eval($extended_env, $cont, $fail);
        },
        $fail
    );
}

##############################
package PScm::SpecialForm::If;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    my $condition = $form->first;
    my $true_branch = $form->rest->first;
    my $false_branch = $form->rest->rest->first;

    $condition->Eval(
        $env,
        cont {
            my ($result, $fail) = @_;
            if ($result->isTrue) {
                $true_branch->Eval($env, $cont, $fail);
            } else {
                $false_branch->Eval($env, $cont, $fail);
            }
        },
        $fail
    );
}

################################
package PScm::SpecialForm::Eval;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;
    $form->first()->Eval(
        $env,
        cont {
            my ($result, $fail) = @_;
            $result->Eval($env, $cont, $fail);
        },
        $fail
    );
}

##################################
package PScm::SpecialForm::Lambda;

use base qw(PScm::SpecialForm);
use PScm::Closure;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    my ($args, $body) = $form->value;
    $cont->Cont(
        PScm::Closure::Function->new($args, $body, $env),
        $fail
    );
}

##################################
package PScm::SpecialForm::CallCC;

use base qw(PScm::SpecialForm);
use PScm::Closure;
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    $form->first->Eval(
        $env,
        cont {
            my ($closure, $fail) = @_;
            $closure->Apply(new PScm::Expr::List($cont),
                $env, $cont);
        },
        $fail
    );
}

##################################
package PScm::SpecialForm::Macro;

use base qw(PScm::SpecialForm);
use PScm::Closure;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;
    my ($args, $body) = $form->value;
    $cont->Cont(
        PScm::Closure::Macro->new($args, $body, $env),
        $fail
    );
}

#################################
package PScm::SpecialForm::Quote;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;
    $form->first->Quote($env, $cont, $fail);
}

###############################
package PScm::SpecialForm::Set;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;
    my ($symbol, $expr) = $form->value;
    my $old_value = $env->LookUpNoError($symbol);
    $expr->Eval(
        $env,
        cont {
            my ($val, $fail) = @_;
            my $result = eval { $env->Assign($symbol, $val) };
            if ($@) {
                $self->Error($@, $env);
            } else {
                $cont->Cont($result, $fail);
            }
        },
        fail {
            $env->Assign($symbol, $old_value);
            $fail->Fail();
        }
    );
}

####################################
package PScm::SpecialForm::Sequence;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;
    if ($form->is_pair) {
        $self->apply_next($form, $env, $cont, $fail);
    } else {
        $cont->Cont($form, $fail);
    }
}


#################################
package PScm::SpecialForm::Begin;

use base qw(PScm::SpecialForm::Sequence);
use PScm::Continuation;

sub apply_next {
    my ($self, $form, $env, $cont, $fail) = @_;

    $form->first->Eval(
        $env,
        cont {
            my ($val, $fail) = @_;
            if ($form->rest->is_pair) {
                $self->apply_next($form->rest, $env, $cont, $fail);
            } else {
                $cont->Cont($val, $fail);
            }
        },
        $fail
    );
}

###############################
package PScm::SpecialForm::And;

use base qw(PScm::SpecialForm::Sequence);
use PScm::Continuation;

sub apply_next {
    my ($self, $form, $env, $cont, $fail) = @_;

    $form->first->Eval(
        $env,
        cont {
            my ($val, $fail) = @_;
            if ($form->rest->is_pair) {
                if ($val->isTrue) {
                    $self->apply_next($form->rest, $env, $cont, $fail);
                } else {
                    $cont->Cont($val, $fail);
                }
            } else {
                $cont->Cont($val, $fail);
            }
        },
        $fail
    );
}

##############################
package PScm::SpecialForm::Or;

use base qw(PScm::SpecialForm::Sequence);
use PScm::Continuation;

sub apply_next {
    my ($self, $form, $env, $cont, $fail) = @_;

    $form->first->Eval(
        $env,
        cont {
            my ($val, $fail) = @_;
            if ($form->rest->is_pair) {
                if ($val->isTrue) {
                    $cont->Cont($val, $fail);
                } else {
                    $self->apply_next($form->rest, $env, $cont, $fail);
                }
            } else {
                $cont->Cont($val, $fail);
            }
        },
        $fail
    );
}

##################################
package PScm::SpecialForm::Define;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;
    my ($symbol, $expr) = $form->value;
    my $old_value = $env->LookUpHere($symbol);

    $expr->Eval(
        $env,
        cont {
            my ($value, $fail) = @_;
            $cont->Cont(
                $env->Define($symbol, $value),
                $fail
            );
        },
        fail {
            if (defined $old_value) {
                $env->Assign($symbol, $old_value);
            } else {
                $env->UnSet($symbol);
            }
            $fail->Fail();
        }
    );
}

#####################################
package PScm::SpecialForm::MakeClass;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    my $parent_expr = $form->first;
    my $fields = $form->rest->first;
    my $methods = $form->rest->rest;

    $parent_expr->Eval(
        $env,
        cont {
            my ($parent_class, $fail) = @_;
            $cont->Cont(
                PScm::Class->new(
                    $parent_class, $fields, $methods, $env
                ),
                $fail
            );
        },
        $fail
    );
}

#################################
package PScm::SpecialForm::Spawn;

use base qw(PScm::SpecialForm);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    PScm->new_thread(
        bounce {
            $cont->Cont(new PScm::Expr::Number(0), $fail);
        }
    );

    $cont->Cont(new PScm::Expr::Number(1), $fail);
}

#################################
package PScm::SpecialForm::Print;

use base qw(PScm::SpecialForm);

use PScm::Continuation;

sub new {
    my ($class, $outfh) = @_;
    bless { outfh => $outfh }, $class;
}

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;
    $form->first->Eval($env,
        cont {
            my ($thing, $fail) = @_;
            $thing->Print(
                $self->{outfh},
                cont {
                    my ($dummy, $fail) = @_;
                    $cont->Cont($thing, $fail)
                },
                $fail
            );
        },
        $fail
    );
}

################################
package PScm::SpecialForm::Exit;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;
    return undef;
}

#################################
package PScm::SpecialForm::Error;

use base qw(PScm::SpecialForm);

use PScm::Continuation;

sub new {
    my ($class, $outfh, $bounce) = @_;
    bless {
        outfh => $outfh,
        bounce => $bounce,
    }, $class;
}

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    $form->first->Eval($env,
        cont {
            my ($msg, $fail) = @_;
            $self->do_error($msg->display_string());
        },
        $fail
    );
}

sub do_error {
    my ($self, $errstr) = @_;
    $errstr =~ s/\n$//;
    $self->{outfh}->print("Error: ", $errstr, "\n");
    return $self->{bounce};
}

###############################
package PScm::SpecialForm::Amb;

use base qw(PScm::SpecialForm);

use PScm::Continuation;

sub Apply {
    my ($self, $choices, $env, $cont, $fail) = @_;
    if ($choices->is_pair) {
        $choices->first->Eval(
            $env,
            $cont,
            fail {
                $self->Apply(
                    $choices->rest,
                    $env,
                    $cont,
                    $fail
                )
            }
        )
    } else {
        $fail->Fail();
    }
}

1;
